"""Unit tests for the render functions."""

import argparse
import unittest

from hypothesis import given, strategies
from pygments.styles import get_all_styles

from next_action import todotxt
from next_action.output import render_next_action, render_arguments


class RenderNextActionTest(unittest.TestCase):
    """Unit tests for the render next action method."""

    def setUp(self):
        """Set up the namespace with default arguments for all unit tests."""
        self.namespace = argparse.Namespace()
        self.namespace.reference = "multiple"
        self.namespace.file = ["todo.txt"]
        self.namespace.style = None
        self.namespace.blocked = False

    def test_reference_always(self):
        """Test that the source filename is added if reference is always."""
        self.namespace.reference = "always"
        self.assertEqual("Todo [todo.txt]", render_next_action([todotxt.Task("Todo", "todo.txt")], [], self.namespace))

    def test_reference_multiple(self):
        """Test that the source filename is added if reference is multiple and there are multiple todo.txt files."""
        self.namespace.file.append("work.txt")
        self.assertEqual(
            "Todo [todo.txt]\nProposal [work.txt]",
            render_next_action([todotxt.Task("Todo", "todo.txt"), todotxt.Task("Proposal", "work.txt")], [],
                               self.namespace))

    def test_reference_never(self):
        """Test that the source filename is not added if reference is never."""
        self.namespace.reference = "never"
        self.namespace.file.append("work.txt")
        self.assertEqual(
            "Todo\nProposal",
            render_next_action([todotxt.Task("Todo", "todo.txt"), todotxt.Task("Proposal", "work.txt")], [],
                               self.namespace))

    def test_blocked(self):
        """Test that the blocked task is rendered."""
        self.namespace.blocked = True
        rinse = todotxt.Task("Rinse before:repeat")
        repeat = todotxt.Task("Repeat id:repeat")
        rinse.add_blocked_task(repeat)
        self.assertEqual(
            "Rinse before:repeat\nblocks:\n- Repeat id:repeat",
            render_next_action([rinse], [], self.namespace))

    def test_blocked_multiple(self):
        """Test that multiple blocked tasks are rendered."""
        self.namespace.blocked = True
        lather = todotxt.Task("Rinse before:repeat before:rinse")
        repeat = todotxt.Task("Repeat id:repeat")
        rinse = todotxt.Task("Rinse id:rinse")
        lather.add_blocked_task(repeat)
        lather.add_blocked_task(rinse)
        self.assertEqual(
            "Rinse before:repeat before:rinse\nblocks:\n- Repeat id:repeat\n- Rinse id:rinse",
            render_next_action([lather], [], self.namespace))

    def test_blocked_recursive(self):
        """Test that the blocked tasks are rendered, recursively."""
        self.namespace.blocked = True
        lather = todotxt.Task("Lather before:rinse")
        rinse = todotxt.Task("Rinse id:rinse before:repeat")
        repeat = todotxt.Task("Repeat id:repeat")
        lather.add_blocked_task(rinse)
        rinse.add_blocked_task(repeat)
        self.assertEqual(
            "Lather before:rinse\nblocks:\n- Rinse id:rinse before:repeat\n  blocks:\n  - Repeat id:repeat",
            render_next_action([lather], [], self.namespace))


class RenderArgumentsTest(unittest.TestCase):
    """Unit tests for the render arguments method."""

    def test_arguments(self):
        """Test that the base arguments are rendered correctly."""
        self.assertEqual("+ -+ --all --blocked --config-file --due --file --help --number --overdue --priority "
                         "--reference --style --time-travel --version -@ -V -a -b -c -d -f -h -n -o -p -r -s -t @",
                         render_arguments("all", todotxt.Tasks()))

    @given(strategies.sampled_from(["__reference", "_r"]))
    def test_reference(self, argument):
        """Test that the reference arguments are rendered correctly."""
        self.assertEqual("always multiple never", render_arguments(argument, todotxt.Tasks()))

    @given(strategies.sampled_from(["__time_travel", "_t"]))
    def test_time_travel(self, argument):
        """Test that the time-travel arguments are rendered correctly."""
        self.assertEqual("tomorrow yesterday Monday Tuesday Wednesday Thursday Friday Saturday Sunday",
                         render_arguments(argument, todotxt.Tasks()))

    @given(strategies.sampled_from(["__style", "_s"]))
    def test_style(self, argument):
        """Test that the style arguments are rendered correctly."""
        self.assertEqual(" ".join(sorted(get_all_styles())), render_arguments(argument, todotxt.Tasks()))

    @given(strategies.sampled_from(["__priority", "_p"]))
    def test_priority_no_prios(self, argument):
        """Test that the priority arguments are rendered correctly."""
        self.assertEqual("", render_arguments(argument, todotxt.Tasks()))

    @given(strategies.sampled_from(["__priority", "_p"]))
    def test_priority(self, argument):
        """Test that the priority arguments are rendered correctly."""
        self.assertEqual("A C",
                         render_arguments(argument, todotxt.Tasks([todotxt.Task("(A) A"), todotxt.Task("(C) C")])))



"""Collection of Todo.txt tasks."""

from typing import List, Set

from .task import Task


class Tasks(List[Task]):
    """Collection of Todo.txt tasks."""

    # pylint: disable=not-an-iterable

    def contexts(self) -> Set[str]:
        """Return the contexts used in the collection of tasks."""
        return set(context for task in self for context in task.contexts())

    def projects(self) -> Set[str]:
        """Return the projects used in the collection of tasks."""
        return set(project for task in self for project in task.projects())

    def priorities(self) -> Set[str]:
        """Return the priorities used in the collection of tasks."""
        return set(str(task.priority()) for task in self if task.priority())



"""Methods for reading, parsing, and validating Next-action configuration files."""

import argparse
import os
import string
import sys
from typing import Callable, Dict, List, Union

from pygments.styles import get_all_styles
import yaml
import cerberus


def read_config_file(filename: str, default_filename: str, error: Callable[[str], None]):
    """Read and parse the configuration file."""
    try:
        with open(os.path.expanduser(filename), "r") as config_file:
            return yaml.safe_load(config_file.read())
    except FileNotFoundError as reason:
        if filename == default_filename:
            # Don't complain if there's no configuration file at the default location
            return dict()  # pragma: no cover-behave
        error(f"can't open file: {reason}")
    except OSError as reason:
        error(f"can't open file: {reason}")
    except yaml.YAMLError as reason:
        error(f"can't parse {filename}: {reason}")


def write_config_file(namespace: argparse.Namespace) -> None:
    """Generate a configuration file on standard out."""
    intro = "# Configuration file for Next-action. Edit the settings below as you like.\n"
    options = dict(file=namespace.file[0] if len(namespace.file) == 1 else namespace.file,
                   reference=namespace.reference, style=namespace.style or "default")
    prefixed_filters = []
    for prefix, filters in (("@", namespace.contexts), ("+", namespace.projects),
                            ("-@", namespace.excluded_contexts), ("-+", namespace.excluded_projects)):
        prefixed_filters.extend([prefix + filter_name for filter_name in filters])
    if prefixed_filters:
        options["filters"] = prefixed_filters
    if namespace.number == sys.maxsize:
        options["all"] = True
    else:
        options["number"] = namespace.number
    if namespace.priority:
        options["priority"] = namespace.priority
    if namespace.blocked:
        options["blocked"] = True
    config = yaml.dump(options, default_flow_style=False)
    sys.stdout.write(intro + config)


def validate_config_file(config, config_filename: str, error: Callable[[str], None]) -> None:
    """Validate the configuration file contents."""
    schema = {
        "file": {
            "type": ["string", "list"],
            "schema": {"type": "string"}
        },
        "number": {
            "type": "integer",
            "min": 1,
            "excludes": "all"
        },
        "all": {
            "type": "boolean",
            "allowed": [True]
        },
        "priority": {
            "type": "string",
            "allowed": [letter for letter in string.ascii_uppercase]
        },
        "filters": {
            "type": ["string", "list"],
            "regex": r"^\-?[@|\+]\S+(\s+\-?[@|\+]\S+)*",
            "schema": {"type": "string", "regex": r"^\-?[@|\+]\S+"}
        },
        "reference": {
            "type": "string",
            "allowed": ["always", "never", "multiple"]
        },
        "style": {
            "type": "string",
            "allowed": sorted(list(get_all_styles()))
        },
        "blocked": {
            "type": "boolean",
            "allowed": [True]
        }
    }
    validator = cerberus.Validator(schema)
    try:
        valid = validator.validate(config)
    except cerberus.validator.DocumentError as reason:
        error(f"{config_filename} is invalid: {reason}")
    if not valid:
        error(f"{config_filename} is invalid: {flatten_errors(validator.errors)}")


def flatten_errors(error_message: Union[Dict, List, str]) -> str:
    """Flatten Cerberus' error messages."""
    def flatten_dict(error_dict: Dict) -> str:
        """Return a string version of the dict."""
        return ", ".join([f"{key}: {flatten_errors(value)}" for key, value in error_dict.items()])

    def flatten_list(error_list: List) -> str:
        """Return a string version of the list."""
        return ", ".join([flatten_errors(item) for item in error_list])

    if isinstance(error_message, dict):
        return flatten_dict(error_message)
    if isinstance(error_message, list):
        return flatten_list(error_message)
    return error_message



"""Unit tests for the reference module in the output package."""

import argparse
import unittest

from next_action import todotxt
from next_action.output import reference


class ReferenceTest(unittest.TestCase):
    """Unit tests for the reference method."""

    def setUp(self):
        """Set up the namespace with default arguments for all unit tests."""
        self.namespace = argparse.Namespace()
        self.namespace.reference = "multiple"

    def test_always(self):
        """Test that the source filename is added."""
        self.namespace.reference = "always"
        self.assertEqual("Todo [todo.txt]", reference(todotxt.Task("Todo", "todo.txt"), self.namespace))

    def test_never(self):
        """Test that the source filename is not added."""
        self.namespace.reference = "never"
        self.assertEqual("Todo", reference(todotxt.Task("Todo", "todo.txt"), self.namespace))

    def test_multiple(self):
        """Test that the source filename is added."""
        self.namespace.file = ["todo.txt", "project.txt"]
        self.assertEqual("Todo [todo.txt]", reference(todotxt.Task("Todo", "todo.txt"), self.namespace))



"""Colorize the output."""

import argparse
import functools

from pygments import lexers, formatters, highlight


@functools.lru_cache(maxsize=None)
def lexer():
    """Return the Todo.txt lexer."""
    return lexers.get_lexer_by_name("todotxt")


@functools.lru_cache(maxsize=None)
def formatter(style: str):
    """Return the formatter for the style."""
    return formatters.get_formatter_by_name("terminal256", style=style)


def colorize(todotxt: str, namespace: argparse.Namespace) -> str:
    """Colorize the todotxt text according to the style."""
    if not namespace.style:
        return todotxt
    result = highlight(todotxt, lexer(), formatter(namespace.style))
    if not todotxt.endswith("\n"):
        # Remove the line break that Pygments added
        result = result[:-1]  # pragma: no cover-behave
    return result



"""Unit tests for the read todo.txt files method."""

import unittest
from unittest.mock import patch, mock_open

from next_action.todotxt import read_todotxt_files


class ReadTodoTxtFilesTest(unittest.TestCase):
    """Unit tests for the read todo.txt files method."""

    @patch("fileinput.open", mock_open(read_data="\nWalk the dog\nx (B) Call mom\n"))
    def test_skip_completed_tasks(self):
        """Test that completed tasks are skipped."""
        tasks = read_todotxt_files(["todo.txt"])
        self.assertEqual(1, len(tasks))
        self.assertEqual("Walk the dog", tasks[0].text)  # pylint: disable=unsubscriptable-object

    @patch("fileinput.open", mock_open(read_data="\nWalk the dog\n   \n\n"))
    def test_skip_empty_lines(self):
        """Test that empty lines are skipped."""
        tasks = read_todotxt_files(["todo.txt"])
        self.assertEqual(1, len(tasks))
        self.assertEqual("Walk the dog", tasks[0].text)  # pylint: disable=unsubscriptable-object

    @patch("fileinput.open", mock_open(read_data="\nBlocked id:1\nBlocking before:1\n"))
    def test_skip_blocked_before(self):
        """Test that tasks blocked with before are skipped."""
        tasks = read_todotxt_files(["todo.txt"])
        self.assertEqual(1, len(tasks))
        self.assertEqual("Blocking before:1", tasks[0].text)  # pylint: disable=unsubscriptable-object

    @patch("fileinput.open", mock_open(read_data="\nBlocking id:1\nBlocked after:1\n"))
    def test_skip_blocked_after(self):
        """Test that tasks blocked with after are skipped."""
        tasks = read_todotxt_files(["todo.txt"])
        self.assertEqual(1, len(tasks))
        self.assertEqual("Blocking id:1", tasks[0].text)  # pylint: disable=unsubscriptable-object

    @patch("fileinput.open", mock_open(read_data="Blocking before:1\n"))
    def test_mssing_before_task(self):
        """Test a mising blocked task."""
        tasks = read_todotxt_files(["todo.txt"])
        self.assertEqual(1, len(tasks))
        self.assertEqual("Blocking before:1", tasks[0].text)  # pylint: disable=unsubscriptable-object

    @patch("fileinput.open", mock_open(read_data="Blocked after:1\n"))
    def test_missing_after_task(self):
        """Test a mising blocked task."""
        tasks = read_todotxt_files(["todo.txt"])
        self.assertEqual(1, len(tasks))
        self.assertEqual("Blocked after:1", tasks[0].text)  # pylint: disable=unsubscriptable-object



"""Unit tests for the todo.txt Task class."""

import datetime
import string
import unittest

from hypothesis import given, strategies

from next_action import todotxt

# pylint: disable=no-member


class TodoTest(unittest.TestCase):
    """Unit tests for the Task class that represents one task item."""

    def test_task_text(self):
        """Test that the task text can be retrieved."""
        self.assertEqual("Todo", todotxt.Task("Todo").text)

    def test_task_repr(self):
        """Test the task repr()."""
        self.assertEqual("Task<Todo>", repr(todotxt.Task("Todo")))


BRACKETS = ["()", "<>", "{}", "[]", "''", '""']


class TodoContextTest(unittest.TestCase):
    """Unit tests for Task contexts."""

    def test_no_context(self):
        """Test that a task without contexts has no contexts."""
        self.assertEqual(set(), todotxt.Task("Todo").contexts())

    def test_one_context(self):
        """Test that the task context can be retrieved."""
        self.assertEqual({"home"}, todotxt.Task("Todo @home").contexts())

    def test_two_contexts(self):
        """Test that multiple task contexts can be retrieved."""
        self.assertEqual({"home", "work"}, todotxt.Task("Todo @home @work").contexts())

    def test_no_space_before_at_sign(self):
        """Test that a context with a space before the @-sign is not considered a context."""
        self.assertEqual(set(), todotxt.Task("Todo@home").contexts())

    def test_at_start_of_line(self):
        """Test that a context at the start of a line is considered a context."""
        self.assertEqual({"Home"}, todotxt.Task("@Home fix leak").contexts())

    @given(strategies.sampled_from(BRACKETS))
    def test_brackets(self, brackets):
        """Test that a brackted context is considered a context."""
        opening, closing = brackets[0], brackets[1]
        self.assertEqual({"home"}, todotxt.Task(f"Clean desk {opening}@home{closing}").contexts())


class TaskProjectTest(unittest.TestCase):
    """Unit tests for task projects."""

    def test_no_projects(self):
        """Test a task without projects."""
        self.assertEqual(set(), todotxt.Task("Todo").projects())

    def test_one_project(self):
        """Test that the task project can be retrieved."""
        self.assertEqual({"projectx"}, todotxt.Task("Todo +projectx").projects())

    def test_two_projects(self):
        """Test that multiple task projects can be retrieved."""
        self.assertEqual({"projectx", "projecty"}, todotxt.Task("Todo +projectx +projecty").projects())

    def test_no_space_before_at_sign(self):
        """Test that a project with a space before the +-sign is not considered a project."""
        self.assertEqual(set(), todotxt.Task("Todo+project").projects())

    def test_at_start_of_line(self):
        """Test that a project at the start of a line is considered a project."""
        self.assertEqual({"Maintenance"}, todotxt.Task("+Maintenance fix leak").projects())

    @given(strategies.sampled_from(BRACKETS))
    def test_brackets(self, brackets):
        """Test that a bracketed project is considered a project."""
        opening, closing = brackets[0], brackets[1]
        self.assertEqual({"GarageSale"}, todotxt.Task(f"Organize {opening}+GarageSale{closing} @home").projects())


class TaskPriorityTest(unittest.TestCase):
    """Unit test for task priority."""

    def test_no_priority(self):
        """Test a task without priority."""
        self.assertEqual(None, todotxt.Task("Todo").priority())

    def test_priorities(self):
        """Test a task with a priority."""
        for priority in string.ascii_uppercase:
            self.assertEqual(priority, todotxt.Task(f"({priority}) Todo").priority())

    def test_faulty_priorities(self):
        """Test that priorities must be one upper case letter."""
        for todo_txt in ("(a) Todo", " (a) Todo", "(A1) Todo", "(A Todo", "Todo (A)", "(A)Todo"):
            self.assertEqual(None, todotxt.Task(todo_txt).priority())

    def test_priority_at_least(self):
        """Test that the task has a minimum priority."""
        self.assertTrue(todotxt.Task("(A) Task").priority_at_least("A"))
        self.assertTrue(todotxt.Task("(A) Task").priority_at_least("B"))
        self.assertFalse(todotxt.Task("(B) Task").priority_at_least("A"))
        self.assertTrue(todotxt.Task("(Z) Task").priority_at_least("Z"))
        self.assertFalse(todotxt.Task("Task").priority_at_least("Z"))

    def test_blocking(self):
        """Test the priority of a task without its own priority.

        Test that the priority of a task without its own priority equals the priority of the task it is
        blocking.
        """
        after = todotxt.Task("(A) After id:1")
        before = todotxt.Task("Before before:1")
        after.set_is_blocked()
        before.add_blocked_task(after)
        self.assertEqual("A", before.priority())

    def test_blocking_multiple(self):
        """Test the priority of a task without its own priority.

        Test that the priority of a task without its own priority equals the highest priority of the tasks it is
        blocking.
        """
        after1 = todotxt.Task("(A) After id:1")
        after2 = todotxt.Task("(B) After after:before")
        before = todotxt.Task("Before before:1 id:before")
        before.add_blocked_task(after1)
        before.add_blocked_task(after2)
        after1.set_is_blocked()
        after2.set_is_blocked()
        self.assertEqual("A", before.priority())


class CreationDateTest(unittest.TestCase):
    """Unit tests for task creation dates.

    Next-action interprets creation dates in the future as threshold, or start date.
    """

    def test_no_creation_date(self):
        """Test that tasks have no creation date by default."""
        self.assertEqual(None, todotxt.Task("Todo").creation_date())

    def test_creation_date(self):
        """Test a valid creation date."""
        self.assertEqual(datetime.date(2018, 1, 2), todotxt.Task("2018-01-02 Todo").creation_date())

    def test_creation_date_after_priority(self):  # pylint:disable=invalid-name
        """Test a valid creation date after the priority."""
        self.assertEqual(datetime.date(2018, 12, 3), todotxt.Task("(B) 2018-12-03 Todo").creation_date())

    def test_invalid_creation_date(self):
        """Test an invalid creation date."""
        self.assertEqual(None, todotxt.Task("2018-14-02 Todo").creation_date())

    def test_no_space_after(self):
        """Test a creation date without a word boundary."""
        self.assertEqual(None, todotxt.Task("2018-10-10Todo").creation_date())

    def test_single_digits(self):
        """Test a creation date with single digits for day and/or month."""
        self.assertEqual(datetime.date(2018, 12, 3), todotxt.Task("(B) 2018-12-3 Todo").creation_date())
        self.assertEqual(datetime.date(2018, 1, 13), todotxt.Task("(B) 2018-1-13 Todo").creation_date())
        self.assertEqual(datetime.date(2018, 1, 1), todotxt.Task("(B) 2018-1-1 Todo").creation_date())

    def test_is_future_task(self):
        """Test that a task with a creation date in the future is a future task."""
        self.assertTrue(todotxt.Task("9999-01-01 Prepare for five-digit years").is_future())
        self.assertFalse(todotxt.Task(f"{datetime.date.today().isoformat()} Todo").is_future())


class ThresholdDateTest(unittest.TestCase):
    """Unit tests for the threshold date of a task.

    The core todo.txt standard only defines creation date, but threshold (t:<date>) seems to be a widely used
    convention.
    """

    def test_no_threshold(self):
        """Test that tasks have no threshold by default."""
        task = todotxt.Task("Todo")
        self.assertEqual(None, task.threshold_date())
        self.assertFalse(task.is_future())

    def test_past_threshold(self):
        """Test a past threshold date."""
        task = todotxt.Task("Todo t:2018-01-02")
        self.assertEqual(datetime.date(2018, 1, 2), task.threshold_date())
        self.assertFalse(task.is_future())

    def test_future_threshold(self):
        """Test a future threshold date."""
        task = todotxt.Task("Todo t:9999-01-01")
        self.assertEqual(datetime.date(9999, 1, 1), task.threshold_date())
        self.assertTrue(task.is_future())

    def test_threshold_today(self):
        """Test a task with threshold today."""
        task = todotxt.Task("Todo t:{0}".format(datetime.date.today().isoformat()))
        self.assertEqual(datetime.date.today(), task.threshold_date())
        self.assertFalse(task.is_future())


class DueDateTest(unittest.TestCase):
    """Unit tests for the due date of tasks."""

    def test_no_due_date(self):
        """Test that tasks have no due date by default."""
        task = todotxt.Task("Todo")
        self.assertEqual(None, task.due_date())
        self.assertFalse(task.is_overdue())

    def test_past_due_date(self):
        """Test a past due date."""
        task = todotxt.Task("Todo due:2018-01-02")
        self.assertEqual(datetime.date(2018, 1, 2), task.due_date())
        self.assertTrue(task.is_overdue())

    def test_future_due_date(self):
        """Test a future due date."""
        task = todotxt.Task("Todo due:9999-01-01")
        self.assertEqual(datetime.date(9999, 1, 1), task.due_date())
        self.assertFalse(task.is_overdue())

    def test_due_today(self):
        """Test a task due today."""
        task = todotxt.Task(f"Todo due:{datetime.date.today().isoformat()}")
        self.assertEqual(datetime.date.today(), task.due_date())
        self.assertFalse(task.is_overdue())

    def test_invalid_date(self):
        """Test an invalid due date."""
        task = todotxt.Task("Todo due:2018-01-32")
        self.assertEqual(None, task.due_date())
        self.assertFalse(task.is_overdue())

    def test_no_space_after(self):
        """Test a due date without a word boundary following it."""
        task = todotxt.Task("Todo due:2018-01-023")
        self.assertEqual(None, task.due_date())
        self.assertFalse(task.is_overdue())

    @given(strategies.sampled_from(["2018-01-1", "2018-1-01", "2018-1-1"]))
    def test_single_digits(self, due_date):
        """Test a due date with single digits for day and/or month."""
        self.assertEqual(datetime.date(2018, 1, 1), todotxt.Task(f"(B) due:{due_date} Todo").due_date())

    def test_is_due(self):
        """Test the is_due method."""
        self.assertTrue(todotxt.Task("due:2018-01-01").is_due(datetime.date.today()))
        self.assertTrue(todotxt.Task("due:2018-01-01").is_due(datetime.date(2018, 1, 1)))
        self.assertTrue(todotxt.Task("9999-01-01 due:2018-01-01").is_due(datetime.date(2018, 1, 1)))
        self.assertFalse(todotxt.Task("due:2018-01-01").is_due(datetime.date(2017, 12, 31)))
        self.assertFalse(todotxt.Task("Without due date").is_due(datetime.date(2017, 12, 31)))

    def test_blocking(self):
        """Test the due date of a task without its own due date.

        Test that the due date of a task without its own due date equals the due date of the task it is blocking.
        """
        after = todotxt.Task("After id:1 due:2018-01-01")
        before = todotxt.Task("Before before:1")
        after.set_is_blocked()
        before.add_blocked_task(after)
        self.assertEqual(datetime.date(2018, 1, 1), before.due_date())

    def test_blocking_multiple(self):
        """Test the due date of a task without its own due date.

        Test that the due date of a task without its own due date equals the earliest due date of the tasks it is
        blocking.
        """
        after1 = todotxt.Task("After id:1 due:2018-10-01")
        after2 = todotxt.Task("After after:before due:2018-01-01")
        before = todotxt.Task("Before before:1 id:before")
        after1.set_is_blocked()
        after2.set_is_blocked()
        before.add_blocked_task(after1)
        before.add_blocked_task(after2)
        self.assertEqual(datetime.date(2018, 1, 1), before.due_date())


class DependenciesTest(unittest.TestCase):
    """Unit tests for dependency relations."""

    before_keys = strategies.sampled_from(["p", "before"])

    @given(strategies.sampled_from(["Todo", "Todo id:", "Todo id:id", "Todo p:", "Todo p:id",
                                    "Todo before:", "Todo before:id", "Todo after:", "Todo after:id"]))
    def test_task_without_dependencies(self, todo_text):
        """Test that a task without dependencies is not blocked."""
        tasks = todotxt.Tasks()
        task = todotxt.Task(todo_text)
        another_task = todotxt.Task("Another task")
        tasks.extend([task, another_task])
        self.assertFalse(task.is_blocked())

    @given(before_keys)
    def test_one_before_another(self, before_key):
        """Test that a task specified to be done before another task blocks the latter."""
        after = todotxt.Task("After id:1")
        before = todotxt.Task(f"Before {before_key}:1")
        after.set_is_blocked()
        before.add_blocked_task(after)
        self.assertTrue(after.is_blocked())

    def test_one_after_another(self):
        """Test that a task specified to be done after another task blocks the first."""
        after = todotxt.Task("After after:1")
        before = todotxt.Task("Before id:1")
        after.set_is_blocked()
        before.add_blocked_task(after)
        self.assertTrue(after.is_blocked())

    @given(before_keys)
    def test_one_before_two(self, before_key):
        """Test that a task that is specified to be done before two other tasks blocks both tasks."""
        before = todotxt.Task(f"Before {before_key}:1 {before_key}:2")
        after1 = todotxt.Task("After id:1")
        after2 = todotxt.Task("After id:1")
        before.add_blocked_task(after1)
        before.add_blocked_task(after2)
        after1.set_is_blocked()
        after2.set_is_blocked()
        self.assertTrue(after1.is_blocked())
        self.assertTrue(after2.is_blocked())

    def test_one_after_two(self):
        """Test that a task that is specified to be done after two other tasks is blocked by both."""
        before1 = todotxt.Task("Before 1 id:1")
        before2 = todotxt.Task("Before 2 id:2")
        after = todotxt.Task("After after:1 after:2")
        before1.add_blocked_task(after)
        before2.add_blocked_task(after)
        after.set_is_blocked()
        self.assertTrue(after.is_blocked())

    @given(before_keys)
    def test_two_before_one(self, before_key):
        """Test that a task that is specified to be done after two other tasks is blocked by both."""
        before1 = todotxt.Task(f"Before 1 {before_key}:1")
        before2 = todotxt.Task(f"Before 2 {before_key}:1")
        after = todotxt.Task("After id:1")
        before1.add_blocked_task(after)
        before2.add_blocked_task(after)
        after.set_is_blocked()
        self.assertTrue(after.is_blocked())

    def test_two_after_one(self):
        """Test that two tasks that are specified to be done after one other task are both blocked."""
        before = todotxt.Task("Before id:before")
        after1 = todotxt.Task("After 1 after:before")
        after2 = todotxt.Task("After 2 after:before")
        before.add_blocked_task(after1)
        before.add_blocked_task(after2)
        after1.set_is_blocked()
        after2.set_is_blocked()
        self.assertTrue(after1.is_blocked())
        self.assertTrue(after2.is_blocked())

    @given(before_keys)
    def test_self_before_self(self, before_key):
        """Test that a task can be blocked by itself.

        This doesn't make sense, but we're not in the business of validating todo.txt files.
        """
        task = todotxt.Task(f"Todo id:1 {before_key}:1")
        task.set_is_blocked()
        task.add_blocked_task(task)
        self.assertTrue(task.is_blocked())

    def test_self_after_self(self):
        """Test that a task can be blocked by itself.

        This doesn't make sense, but we're not in the business of validating todo.txt files.
        """
        task = todotxt.Task("Todo id:1 after:1")
        task.set_is_blocked()
        task.add_blocked_task(task)
        self.assertTrue(task.is_blocked())

    @given(before_keys)
    def test_self_before_self_indirect(self, before_key):
        """Test that a task can be blocked by a second task that is blocked by the first task.

        This doesn't make sense, but we're not in the business of validating todo.txt files.
        """
        task1 = todotxt.Task(f"Task 1 id:1 {before_key}:2")
        task2 = todotxt.Task(f"Task 2 id:2 {before_key}:1")
        task1.set_is_blocked()
        task2.set_is_blocked()
        task1.add_blocked_task(task2)
        task2.add_blocked_task(task1)
        self.assertTrue(task1.is_blocked())

    def test_self_after_self_indirect(self):
        """Test that a task can be blocked by a second task that is blocked by the first task.

        This doesn't make sense, but we're not in the business of validating todo.txt files.
        """
        task1 = todotxt.Task("Task 1 id:1 after:2")
        task2 = todotxt.Task("Task 2 id:2 after:1")
        task1.set_is_blocked()
        task2.set_is_blocked()
        task1.add_blocked_task(task2)
        task2.add_blocked_task(task1)
        self.assertTrue(task1.is_blocked())



"""Package for dealing with the todo.txt format."""

import fileinput
import os
from typing import cast, List, Sequence

from .task import Task
from .tasks import Tasks


def uncompleted_task_on(line: str) -> bool:
    """Return whether there's an uncompleted task on the line."""
    return bool(line.strip() and not line.startswith("x "))


def unblocked_tasks(tasks: Sequence[Task]) -> Sequence[Task]:
    """Set the blocked status of the tasks and return only the unblocked tasks."""
    task_by_id = {task.task_id(): task for task in tasks if task.task_id()}
    for task in tasks:
        for parent_id in task.parent_ids():
            parent_task = task_by_id.get(parent_id)
            if parent_task:
                parent_task.set_is_blocked()
                task.add_blocked_task(parent_task)
        for child_id in task.child_ids():
            child_task = task_by_id.get(child_id)
            if child_task:
                task.set_is_blocked()
                child_task.add_blocked_task(task)
    return [task for task in tasks if not task.is_blocked()]


def read_todotxt_files(filenames: List[str]) -> Tasks:
    """Read tasks from the Todo.txt files."""
    filenames = [os.path.expanduser(filename) for filename in filenames]
    with cast(fileinput.FileInput, fileinput.input(filenames)) as todotxt_file:
        tasks = [Task(line.strip(), todotxt_file.filename()) for line in todotxt_file if uncompleted_task_on(line)]
    return Tasks(unblocked_tasks(tasks))



"""Package for parsing command line arguments."""

import argparse
import os
import shutil
from typing import Tuple

from .parser import NextActionArgumentParser


def parse_arguments(version: str = "?") -> Tuple[argparse.ArgumentParser, argparse.Namespace]:
    """Build the argument parser and parse the command line arguments."""
    # Ensure that the help info is printed using all columns available
    os.environ['COLUMNS'] = str(shutil.get_terminal_size().columns)
    parser = NextActionArgumentParser(version)
    return parser, parser.parse_args()



"""Reference the next actions with their source filename."""

import argparse

from .. import todotxt


def reference(task: todotxt.Task, namespace: argparse.Namespace) -> str:
    """Decorate the next action with a reference to its source filename."""
    rendered_task = task.text
    if namespace.reference == "always" or namespace.reference == "multiple" and len(namespace.file) > 1:
        rendered_task += f" [{task.filename}]"
    return rendered_task



"""Fixtures for unit tests."""

import argparse
import unittest
import sys


class TestCaseWithNamespace(unittest.TestCase):
    """Test case for unit tests that need a argparse.Namespace."""

    def setUp(self):
        """Set up the namespace with default arguments for all unit tests."""
        self.namespace = argparse.Namespace()
        self.namespace.contexts = set()
        self.namespace.projects = set()
        self.namespace.excluded_contexts = set()
        self.namespace.excluded_projects = set()
        self.namespace.overdue = False
        self.namespace.due = None
        self.namespace.priority = None
        self.namespace.time_travel = None
        self.namespace.number = sys.maxsize



"""Package for formatting output."""

import argparse
from typing import Iterable

from pygments.styles import get_all_styles

from .. import todotxt, arguments
from .color import colorize
from .reference import reference
from .warning import invalid_arguments


def render_blocked_tasks(task: todotxt.Task, namespace: argparse.Namespace, level: int = 0) -> str:
    """Render the tasks blocked by the task, if any."""
    rendered_blocked_tasks = ""
    if namespace.blocked:
        blocked_tasks = task.blocked_tasks()
        if blocked_tasks:
            rendered_blocked_tasks += "\n" + level * "  " + "blocks:"
            for blocked_task in blocked_tasks:
                rendered_blocked_tasks += "\n" + render_task(blocked_task, namespace, level + 1)
    return rendered_blocked_tasks


def render_task(task: todotxt.Task, namespace: argparse.Namespace, level: int = 0) -> str:
    """Render one task."""
    indent = (level - 1) * "  " + "- " if level else ""
    rendered_task = colorize(reference(task, namespace), namespace)
    rendered_blocked_tasks = render_blocked_tasks(task, namespace, level)
    return indent + rendered_task + rendered_blocked_tasks


def render_tasks(tasks: todotxt.Tasks, namespace: argparse.Namespace) -> str:
    """Render the tasks using the options in the namespace."""
    return "\n".join(render_task(task, namespace) for task in tasks)


def render_nothing_todo(tasks: todotxt.Tasks, namespace: argparse.Namespace):
    """Tell the user there's nothing to do and warn about invalid arguments, if any."""
    warning = invalid_arguments(namespace, tasks)
    return "Nothing to do!" + (warning if warning else " 😴")


def render_next_action(next_actions: todotxt.Tasks, tasks: todotxt.Tasks, namespace: argparse.Namespace) -> str:
    """Render the next action(s) or, if there are none, tell the user there's nothing to do."""
    return render_tasks(next_actions, namespace) if next_actions else render_nothing_todo(tasks, namespace)


def render_arguments(argument_type: str, tasks: todotxt.Tasks) -> str:
    """Return the argument for tab completion."""
    argument_type = argument_type.replace("_", "-")  # Undo escaping
    argument_values: Iterable[str]
    if argument_type in ("--time-travel", "-t"):
        return "tomorrow yesterday Monday Tuesday Wednesday Thursday Friday Saturday Sunday"
    if argument_type in ("--reference", "-r"):
        argument_values = arguments.parser.REFERENCE_CHOICES
    elif argument_type in ("--style", "-s"):
        argument_values = get_all_styles()
    elif argument_type in ("--priority", "-p"):
        argument_values = tasks.priorities()
    elif argument_type in ("@", "-@"):
        argument_values = (f"{argument_type}{context}" for context in tasks.contexts())
    elif argument_type in ("+", "-+"):
        argument_values = (f"{argument_type}{project}" for project in tasks.projects())
    else:
        argument_values = arguments.parser.ARGUMENTS
    return " ".join(sorted(argument_values))



"""Class that represents one task (i.e. one line) from a todo.txt file."""

import datetime
import functools
import re
import typing
from typing import List, Optional, Sequence, Set


class Task:
    """A task from a line in a todo.txt file."""

    iso_date_reg_exp = r"(\d{4})-(\d{1,2})-(\d{1,2})"

    def __init__(self, todo_txt: str, filename: str = "") -> None:
        """Initialise the task with its Todo.txt text string and originating filename."""
        self.text = todo_txt
        self.filename = filename
        self.__is_blocked = False
        self.__blocked_tasks: List["Task"] = []

    def __repr__(self) -> str:
        """Return a text representation of the task."""
        return f"{self.__class__.__name__}<{self.text}>"

    @functools.lru_cache(maxsize=None)
    def contexts(self) -> Set[str]:
        """Return the contexts of the task."""
        return self.__prefixed_items("@")

    @functools.lru_cache(maxsize=None)
    def projects(self) -> Set[str]:
        """Return the projects of the task."""
        return self.__prefixed_items(r"\+")

    @functools.lru_cache(maxsize=None)
    def priority(self) -> Optional[str]:
        """Return the priority of the task."""
        match = re.match(r"\(([A-Z])\) ", self.text)
        priorities: List[Optional[str]] = [match.group(1)] if match else []
        priorities.extend([blocked_task.priority() for blocked_task in self.blocked_tasks()])
        return min(priorities, default=None, key=lambda priority: priority or "ZZZ")

    def priority_at_least(self, min_priority: str) -> bool:
        """Return whether the priority of task is at least the given priority."""
        priority = self.priority()
        if priority:
            return priority <= min_priority
        return False

    @functools.lru_cache(maxsize=None)
    def creation_date(self) -> Optional[datetime.date]:
        """Return the creation date of the task."""
        match = re.match(fr"(?:\([A-Z]\) )?{self.iso_date_reg_exp}\b", self.text)
        return self.__create_date(match)

    @functools.lru_cache(maxsize=None)
    def threshold_date(self) -> Optional[datetime.date]:
        """Return the threshold date of the task."""
        return self.__find_keyed_date("t")

    @functools.lru_cache(maxsize=None)
    def due_date(self) -> Optional[datetime.date]:
        """Return the due date of the task."""
        due_dates = [self.__find_keyed_date("due")]
        due_dates.extend([blocked_task.due_date() for blocked_task in self.blocked_tasks()])
        return min(due_dates, default=None, key=lambda due_date: due_date or datetime.date.max)

    def is_due(self, due_date: datetime.date) -> bool:
        """Return whether the task is due on or before the given due date."""
        task_due_date = self.due_date()
        return task_due_date <= due_date if task_due_date else False

    def is_future(self, today: datetime.date = None) -> bool:
        """Return whether the task is a future task, i.e. has a creation or threshold date in the future."""
        today = today or datetime.date.today()
        creation_date = self.creation_date()
        if creation_date:
            return creation_date > today
        threshold_date = self.threshold_date()
        if threshold_date:
            return threshold_date > today
        return False

    def is_overdue(self, today: datetime.date = None) -> bool:
        """Return whether the task is overdue, i.e. whether it has a due date in the past."""
        today = today or datetime.date.today()
        due_date = self.due_date()
        return due_date < today if due_date else False

    def is_blocked(self) -> bool:
        """Return whether a task is blocked, i.e. whether it has (uncompleted) child tasks."""
        return self.__is_blocked

    def set_is_blocked(self) -> None:
        """Set the blocked status."""
        self.__is_blocked = True

    def blocked_tasks(self) -> Sequence["Task"]:
        """Return the tasks this task is blocking."""
        return self.__blocked_tasks

    def add_blocked_task(self, task: "Task") -> None:
        """Add the task to the blocked tasks."""
        self.__blocked_tasks.append(task)

    @functools.lru_cache(maxsize=None)
    def child_ids(self) -> Set[str]:
        """Return the ids of the child tasks."""
        return {match.group(1) for match in re.finditer(r"\bafter:(\S+)\b", self.text)}

    @functools.lru_cache(maxsize=None)
    def parent_ids(self) -> Set[str]:
        """Return the ids of the parent tasks."""
        return {match.group(2) for match in re.finditer(r"\b(p|before):(\S+)\b", self.text)}

    @functools.lru_cache(maxsize=None)
    def task_id(self) -> str:
        """Return the id of the task."""
        match = re.search(r"\bid:(\S+)\b", self.text)
        return match.group(1) if match else ""

    def __prefixed_items(self, prefix: str) -> Set[str]:
        """Return the prefixed items in the task."""
        return {match.group(2) for match in re.finditer(fr"(^|\W){prefix}(\w+)", self.text)}

    def __find_keyed_date(self, key: str) -> Optional[datetime.date]:
        """Find a key:value pair with the supplied key where the value is a date."""
        match = re.search(fr"\b{key}:{self.iso_date_reg_exp}\b", self.text)
        return self.__create_date(match)

    @staticmethod
    def __create_date(match: Optional[typing.Match[str]]) -> Optional[datetime.date]:
        """Create a date from the match, if possible."""
        if match:
            try:
                return datetime.date(*(int(group) for group in match.groups()))
            except ValueError:
                pass
        return None



"""Unit tests for the argument parsing classes."""

import datetime
import os
import sys
import textwrap
import unittest
from unittest.mock import call, mock_open, patch

from next_action.arguments import config, parse_arguments


USAGE_MESSAGE = textwrap.fill(
    "Usage: next-action [-h] [-V] [-c [<config.cfg>] | -w] [-f <todo.txt> ...] [-t [<date>]] [-b] [-r <ref>] "
    "[-s [<style>]] [-a | -n <number>] [-d [<due date>] | -o] [-p [<priority>]] [--] [<context|project> ...]", 120) + \
    "\n"


class ParserTestCase(unittest.TestCase):
    """Base class for the parser unit tests."""

    def setUp(self):
        """Set up the terminal width for all unit tests."""
        os.environ['COLUMNS'] = "120"  # Fake that the terminal is wide enough.


@patch.object(config, "open", mock_open(read_data=""))
class NoArgumentTest(ParserTestCase):
    """Unit tests for the argument parser, without arguments."""

    @patch.object(sys, "argv", ["next-action"])
    def test_filters(self):
        """Test that the argument parser returns no filters if the user doesn't pass one."""
        self.assertEqual(set(), parse_arguments()[1].contexts)
        self.assertEqual(set(), parse_arguments()[1].projects)
        self.assertEqual(set(), parse_arguments()[1].excluded_contexts)
        self.assertEqual(set(), parse_arguments()[1].excluded_projects)

    @patch.object(sys, "argv", ["next-action"])
    def test_filename(self):
        """Test that the argument parser returns the default filename if the user doesn't pass one."""
        self.assertEqual(["~/todo.txt"], parse_arguments()[1].file)

    @patch.object(sys, "argv", ["next-action"])
    def test_style(self):
        """Test that the argument parser returns the default style if the user doesn't pass one."""
        self.assertEqual(None, parse_arguments()[1].style)


@patch.object(config, "open", mock_open(read_data=""))
class FilenameTest(ParserTestCase):
    """Unit tests for the --filename argument."""

    @patch.object(sys, "argv", ["next-action", "-f", "my_todo.txt"])
    def test_filename_argument(self):
        """Test that the argument parser accepts a filename."""
        self.assertEqual(["my_todo.txt"], parse_arguments()[1].file)

    @patch.object(sys, "argv", ["next-action", "-f", "~/my_todo.txt"])
    def test_home_folder_argument(self):
        """Test that the argument parser accepts a filename with a tilde."""
        self.assertEqual(["~/my_todo.txt"], parse_arguments()[1].file)

    @patch.object(sys, "argv", ["next-action", "--file", "my_other_todo.txt"])
    def test_long_filename_argument(self):
        """Test that the argument parser accepts a filename."""
        self.assertEqual(["my_other_todo.txt"], parse_arguments()[1].file)

    @patch.object(sys, "argv", ["next-action", "-f", "~/todo.txt"])
    def test_add_default_filename(self):
        """Test that adding the default filename doesn't get it included twice."""
        self.assertEqual(["~/todo.txt"], parse_arguments()[1].file)

    @patch.object(sys, "argv", ["next-action", "-f", "todo.txt", "-f", "other.txt"])
    def test_default_and_non_default(self):
        """Test that adding the default filename and another filename gets both included."""
        self.assertEqual(["todo.txt", "other.txt"], parse_arguments()[1].file)

    @patch.object(sys, "argv", ["next-action", "-f", "other.txt", "-f", "other.txt"])
    def test_add_filename_twice(self):
        """Test that adding the same filename twice includes it only once."""
        self.assertEqual(["other.txt"], parse_arguments()[1].file)


@patch.object(config, "open", mock_open(read_data=""))
class FilterArgumentTest(ParserTestCase):
    """Unit tests for the @object and +project filter arguments."""

    @patch.object(sys, "argv", ["next-action", "@home"])
    def test_one_context(self):
        """Test that the argument parser returns the context if the user passes one."""
        self.assertEqual({"home"}, parse_arguments()[1].contexts)

    @patch.object(sys, "argv", ["next-action", "@home", "@work"])
    def test_multiple_contexts(self):
        """Test that the argument parser returns all contexts if the user passes multiple contexts."""
        self.assertEqual({"home", "work"}, parse_arguments()[1].contexts)

    @patch.object(sys, "argv", ["next-action", "@"])
    @patch.object(sys.stderr, "write")
    def test_empty_context(self, mock_stderr_write):
        """Test that the argument parser exits if the context is empty."""
        self.assertRaises(SystemExit, parse_arguments)
        self.assertEqual([call(USAGE_MESSAGE),
                          call("next-action: error: argument <context|project>: context name missing\n")],
                         mock_stderr_write.call_args_list)

    @patch.object(sys, "argv", ["next-action", "-@home"])
    def test_exclude_context(self):
        """Test that contexts can be excluded."""
        self.assertEqual({"home"}, parse_arguments()[1].excluded_contexts)

    @patch.object(sys, "argv", ["next-action", "@home", "-@home"])
    @patch.object(sys.stderr, "write")
    def test_include_exclude_context(self, mock_stderr_write):
        """Test that contexts cannot be included and excluded."""
        self.assertRaises(SystemExit, parse_arguments)
        self.assertEqual([call(USAGE_MESSAGE),
                          call("next-action: error: @home is both included and excluded\n")],
                         mock_stderr_write.call_args_list)

    @patch.object(sys, "argv", ["next-action", "-^"])
    @patch.object(sys.stderr, "write")
    def test_invalid_extra_argument(self, mock_stderr_write):
        """Test that the argument parser exits if the extra argument is invalid."""
        self.assertRaises(SystemExit, parse_arguments)
        self.assertEqual([call(USAGE_MESSAGE), call("next-action: error: unrecognized argument: -^\n")],
                         mock_stderr_write.call_args_list)

    @patch.object(sys, "argv", ["next-action", "+DogHouse"])
    def test_one_project(self):
        """Test that the argument parser returns the project if the user passes one."""
        self.assertEqual({"DogHouse"}, parse_arguments()[1].projects)

    @patch.object(sys, "argv", ["next-action", "+DogHouse", "+PaintHouse"])
    def test_multiple_projects(self):
        """Test that the argument parser returns the projects if the user passes multiple projects."""
        self.assertEqual({"DogHouse", "PaintHouse"}, parse_arguments()[1].projects)

    @patch.object(sys, "argv", ["next-action", "+"])
    @patch.object(sys.stderr, "write")
    def test_empty_project(self, mock_stderr_write):
        """Test that the argument parser exits if the project is empty."""
        self.assertRaises(SystemExit, parse_arguments)
        self.assertEqual([call(USAGE_MESSAGE),
                          call("next-action: error: argument <context|project>: project name missing\n")],
                         mock_stderr_write.call_args_list)

    @patch.object(sys, "argv", ["next-action", "-+DogHouse"])
    def test_exclude_project(self):
        """Test that projects can be excluded."""
        self.assertEqual({"DogHouse"}, parse_arguments()[1].excluded_projects)

    @patch.object(sys, "argv", ["next-action", "+DogHouse", "-+DogHouse"])
    @patch.object(sys.stderr, "write")
    def test_include_exclude_project(self, mock_stderr_write):
        """Test that projects cannot be included and excluded."""
        self.assertRaises(SystemExit, parse_arguments)
        self.assertEqual([call(USAGE_MESSAGE),
                          call("next-action: error: +DogHouse is both included and excluded\n")],
                         mock_stderr_write.call_args_list)

    @patch.object(sys, "argv", ["next-action", "-+"])
    @patch.object(sys.stderr, "write")
    def test_empty_excluded_project(self, mock_stderr_write):
        """Test that the argument parser exits if the project is empty."""
        self.assertRaises(SystemExit, parse_arguments)
        self.assertEqual([call(USAGE_MESSAGE),
                          call("next-action: error: argument <context|project>: project name missing\n")],
                         mock_stderr_write.call_args_list)

    @patch.object(sys, "argv", ["next-action", "+DogHouse", "@home", "+PaintHouse", "@weekend"])
    def test_contexts_and_projects(self):
        """Test that the argument parser returns the contexts and the projects, even when mixed."""
        self.assertEqual({"home", "weekend"}, parse_arguments()[1].contexts)
        self.assertEqual({"DogHouse", "PaintHouse"}, parse_arguments()[1].projects)

    @patch.object(sys, "argv", ["next-action", "@home", "-@work", "+PaintHouse"])
    def test_project_after_excluded(self):
        """Test project after excluded context."""
        namespace = parse_arguments()[1]
        self.assertEqual({"home"}, namespace.contexts)
        self.assertEqual({"work"}, namespace.excluded_contexts)
        self.assertEqual({"PaintHouse"}, namespace.projects)

    @patch.object(sys, "argv", ["next-action", "home"])
    @patch.object(sys.stderr, "write")
    def test_faulty_option(self, mock_stderr_write):
        """Test that the argument parser exits if the option is faulty."""
        self.assertRaises(SystemExit, parse_arguments)
        self.assertEqual([call(USAGE_MESSAGE),
                          call("next-action: error: argument <context|project>: unrecognized argument: home\n")],
                         mock_stderr_write.call_args_list)


@patch.object(config, "open", mock_open(read_data=""))
class NumberTest(ParserTestCase):
    """Unit tests for the --number and --all arguments."""

    @patch.object(sys, "argv", ["next-action"])
    def test_default_number(self):
        """Test that the argument parser has a default number of actions to return."""
        self.assertEqual(1, parse_arguments()[1].number)

    @patch.object(sys, "argv", ["next-action", "--number", "3"])
    def test_number(self):
        """Test that a number of actions to be shown can be passed."""
        self.assertEqual(3, parse_arguments()[1].number)

    @patch.object(sys, "argv", ["next-action", "--number", "not_a_number"])
    @patch.object(sys.stderr, "write")
    def test_faulty_number(self, mock_stderr_write):
        """Test that the argument parser exits if the option is faulty."""
        self.assertRaises(SystemExit, parse_arguments)
        self.assertEqual([call(USAGE_MESSAGE),
                          call("next-action: error: argument -n/--number: invalid number: not_a_number\n")],
                         mock_stderr_write.call_args_list)

    @patch.object(sys, "argv", ["next-action", "--number", "-1"])
    @patch.object(sys.stderr, "write")
    def test_negative_number(self, mock_stderr_write):
        """Test that the argument parser exits if the option is faulty."""
        self.assertRaises(SystemExit, parse_arguments)
        self.assertEqual([call(USAGE_MESSAGE),
                          call("next-action: error: argument -n/--number: invalid number: -1\n")],
                         mock_stderr_write.call_args_list)

    @patch.object(sys, "argv", ["next-action", "--all"])
    def test_all_actions(self):
        """Test that --all option also sets the number of actions to show to a very big number."""
        self.assertEqual(sys.maxsize, parse_arguments()[1].number)

    @patch.object(sys, "argv", ["next-action", "--all", "--number", "3"])
    @patch.object(sys.stderr, "write")
    def test_all_and_number(self, mock_stderr_write):
        """Test that the argument parser exits if the both --all and --number are used."""
        self.assertRaises(SystemExit, parse_arguments)
        self.assertEqual([call(USAGE_MESSAGE),
                          call("next-action: error: argument -n/--number: not allowed with argument -a/--all\n")],
                         mock_stderr_write.call_args_list)


@patch.object(config, "open", mock_open(read_data=""))
class DueDateTest(ParserTestCase):
    """Unit tests for the --due option."""

    @patch.object(sys, "argv", ["next-action"])
    def test_default(self):
        """Test that the default value for due date is None."""
        self.assertEqual(None, parse_arguments()[1].due)

    @patch.object(sys, "argv", ["next-action", "--due"])
    def test_no_due_date(self):
        """Test that the due date is the max date if the user doesn't specify a date."""
        self.assertEqual(datetime.date.max, parse_arguments()[1].due)

    @patch.object(sys, "argv", ["next-action", "--due", "2018-01-01"])
    def test_due_date(self):
        """Test that the default value for due date is None."""
        self.assertEqual(datetime.date(2018, 1, 1), parse_arguments()[1].due)

    @patch.object(sys, "argv", ["next-action", "--due", "not_a_date"])
    @patch.object(sys.stderr, "write")
    def test_faulty_date(self, mock_stderr_write):
        """Test that the argument parser exits if the option is faulty."""
        self.assertRaises(SystemExit, parse_arguments)
        self.assertEqual([call(USAGE_MESSAGE),
                          call("next-action: error: argument -d/--due: invalid date: not_a_date\n")],
                         mock_stderr_write.call_args_list)

    @patch.object(sys, "argv", ["next-action", "--due", "2019-02-30"])
    @patch.object(sys.stderr, "write")
    def test_invalid_date(self, mock_stderr_write):
        """Test that the argument parser exits if the option is invalid."""
        self.assertRaises(SystemExit, parse_arguments)
        self.assertEqual([call(USAGE_MESSAGE),
                          call("next-action: error: argument -d/--due: invalid date: 2019-02-30\n")],
                         mock_stderr_write.call_args_list)

    @patch.object(sys, "argv", ["next-action", "--due", "2019-02-15-12"])
    @patch.object(sys.stderr, "write")
    def test_too_long(self, mock_stderr_write):
        """Test that the argument parser exits if the option is invalid."""
        self.assertRaises(SystemExit, parse_arguments)
        self.assertEqual([call(USAGE_MESSAGE),
                          call("next-action: error: argument -d/--due: invalid date: 2019-02-15-12\n")],
                         mock_stderr_write.call_args_list)

    @patch.object(sys, "argv", ["next-action", "--due", "Extra 2019-02-15"])
    @patch.object(sys.stderr, "write")
    def test_extra_tokens(self, mock_stderr_write):
        """Test that the argument parser exits if the option is invalid."""
        self.assertRaises(SystemExit, parse_arguments)
        self.assertEqual([call(USAGE_MESSAGE),
                          call("next-action: error: argument -d/--due: invalid date: Extra 2019-02-15\n")],
                         mock_stderr_write.call_args_list)


@patch.object(config, "open", mock_open(read_data=""))
class ReferenceTest(ParserTestCase):
    """Unit tests for the --reference option."""

    @patch.object(sys, "argv", ["next-action"])
    def test_default(self):
        """Test that the default value for the reference value is multiple."""
        self.assertEqual("multiple", parse_arguments()[1].reference)

    @patch.object(sys, "argv", ["next-action", "--reference", "multiple"])
    def test_multiple(self):
        """Test that the value can be set to multiple, even though it's the default."""
        self.assertEqual("multiple", parse_arguments()[1].reference)

    @patch.object(sys, "argv", ["next-action", "--reference", "always"])
    def test_always(self):
        """Test that the value can be set to always."""
        self.assertEqual("always", parse_arguments()[1].reference)

    @patch.object(sys, "argv", ["next-action", "--reference", "never"])
    def test_never(self):
        """Test that the value can be set to never."""
        self.assertEqual("never", parse_arguments()[1].reference)

    @patch.object(sys, "argv", ["next-action", "--reference", "not_an_option"])
    @patch.object(sys.stderr, "write")
    def test_faulty_date(self, mock_stderr_write):
        """Test that the argument parser exits if the option is faulty."""
        self.assertRaises(SystemExit, parse_arguments)
        self.assertEqual([call(USAGE_MESSAGE),
                          call("next-action: error: argument -r/--reference: invalid choice: 'not_an_option' "
                               "(choose from 'always', 'never', 'multiple')\n")],
                         mock_stderr_write.call_args_list)


@patch.object(config, "open", mock_open(read_data=""))
class TimeTravelTest(ParserTestCase):
    """Unit tests for the --time-travel option."""

    @patch.object(sys, "argv", ["next-action"])
    def test_default(self):
        """Test that the value for time travel date is None if the option is not present."""
        self.assertEqual(None, parse_arguments()[1].time_travel)

    @patch.object(sys, "argv", ["next-action", "--time-travel"])
    def test_default_value(self):
        """Test that the value for time travel date is tomorrow if the option has no argument."""
        tomorrow = datetime.date.today() + datetime.timedelta(days=1)
        self.assertEqual(tomorrow, parse_arguments()[1].time_travel)

    @patch.object(sys, "argv", ["next-action", "--time-travel", "2018-8-29"])
    def test_specific_date(self):
        """Test that the value for time travel date is the date that is given."""
        self.assertEqual(datetime.date(2018, 8, 29), parse_arguments()[1].time_travel)

    @patch.object(sys, "argv", ["next-action", "--time-travel", "tomorrow", "--due", "today"])
    def test_due_date_is_corrected(self):
        """Test that the value for due date is adapted when time traveling."""
        self.assertEqual(datetime.date.today() + datetime.timedelta(days=1), parse_arguments()[1].due)



"""Main Next-action package."""

from .arguments import parse_arguments
from .pick_action import next_actions
from .output import render_next_action, render_arguments
from .todotxt import read_todotxt_files


__title__ = "next-action"
__version__ = "1.9.0"


def next_action() -> None:
    """Entry point for the command-line interface.

    Basic recipe:
    1) parse command-line arguments,
    2) read todo.txt file(s),
    3) determine the next action(s) and,
    4) display them.
    """
    parser, namespace = parse_arguments(__version__)
    try:
        tasks = read_todotxt_files(namespace.file)
    except OSError as reason:
        parser.error(f"can't open file: {reason}")
    if namespace.list_arguments:
        print(render_arguments(namespace.list_arguments, tasks))
    else:
        actions = next_actions(tasks, namespace)
        print(render_next_action(actions, tasks, namespace))



"""Parser for the command line arguments."""

import argparse
import datetime
import shutil
import string
import sys
import textwrap
from typing import cast, List, Set, Tuple

from pygments.styles import get_all_styles

from .config import read_config_file, write_config_file, validate_config_file


ARGUMENTS = ("@", "+", "-@", "-+", "-a", "--all", "-b", "--blocked", "-c", "--config-file", "-d", "--due",
             "-f", "--file", "-h", "--help", "-n", "--number", "-o", "--overdue", "-p", "--priority", "-r",
             "--reference", "-s", "--style", "-t", "--time-travel", "-V", "--version")
REFERENCE_CHOICES = ("always", "never", "multiple")


class NextActionArgumentParser(argparse.ArgumentParser):
    """Command-line argument parser for Next-action."""

    def __init__(self, version: str = "?") -> None:
        """Initialize the parser."""
        super().__init__(
            usage=textwrap.fill("next-action [-h] [-V] [-c [<config.cfg>] | -w] [-f <todo.txt> ...] "
                                "[-t [<date>]] [-b] [-r <ref>] [-s [<style>]] [-a | -n <number>] "
                                "[-d [<due date>] | -o] [-p [<priority>]] [--] [<context|project> ...]",
                                width=shutil.get_terminal_size().columns - len("usage: ")),
            description="Show the next action in your todo.txt. The next action is selected from the tasks in the "
                        "todo.txt file based on task properties such as priority, due date, and creation date. Limit "
                        "the tasks from which the next action is selected by specifying contexts the tasks must have "
                        "and/or projects the tasks must belong to.",
            epilog="Use -- to separate options with optional arguments from contexts and projects, in order to handle "
                   "cases where a context or project is mistaken for an argument to an option.",
            formatter_class=CapitalisedHelpFormatter)
        self.__default_filenames = ["~/todo.txt"]
        self.add_optional_arguments(version)
        self.add_configuration_options()
        self.add_input_options()
        self.add_output_options()
        self.add_number_options()
        self.add_filter_arguments()

    def add_optional_arguments(self, version: str) -> None:
        """Add the optional arguments to the parser."""
        self._optionals.title = self._optionals.title.capitalize() if self._optionals.title else None
        self.add_argument(
            "-V", "--version", action="version", version=f"%(prog)s {version}")

    def add_configuration_options(self) -> None:
        """Add the configuration options to the parser."""
        config_group = self.add_argument_group("Configuration options")
        config_group.add_argument(
            "-c", "--config-file", metavar="<config.cfg>", type=str, default="~/.next-action.cfg", nargs="?",
            help="filename of configuration file to read (default: %(default)s); omit filename to not read any "
                 "configuration file")
        config_group.add_argument(
            "-w", "--write-config-file", help="generate a sample configuration file and exit", action="store_true")

    def add_input_options(self) -> None:
        """Add the input options to the parser."""
        input_group = self.add_argument_group("Input options")
        input_group.add_argument(
            "-f", "--file", action="append", metavar="<todo.txt>", default=self.__default_filenames[:], type=str,
            help="filename of todo.txt file to read; can be '-' to read from standard input; argument can be "
                 "repeated to read tasks from multiple todo.txt files (default: ~/todo.txt)")
        input_group.add_argument(
            "-t", "--time-travel", metavar="<date>", type=date_type, nargs="?", const="tomorrow",
            help="time travel to the given date and show the next action(s) at that date (default: tomorrow)")

    def add_output_options(self) -> None:
        """Add the output/styling options to the parser."""
        output_group = self.add_argument_group("Output options")
        output_group.add_argument(
            "-b", "--blocked", help="show the tasks blocked by the next action, if any (default: %(default)s)",
            action="store_true")
        output_group.add_argument(
            "-r", "--reference", choices=REFERENCE_CHOICES, default="multiple",
            help="reference next actions with the name of their todo.txt file (default: when reading multiple "
                 "todo.txt files)")
        styles = sorted(list(get_all_styles()))
        output_group.add_argument(
            "-s", "--style", metavar="<style>", choices=styles, default=None, nargs="?",
            help=f"colorize the output; available styles: {', '.join(styles)} (default: %(default)s)")

    def add_number_options(self) -> None:
        """Add the number options to the parser."""
        number_group = self.add_argument_group("Show multiple next actions")
        number = number_group.add_mutually_exclusive_group()
        number.add_argument(
            "-a", "--all", default=1, action="store_const", dest="number", const=sys.maxsize,
            help="show all next actions")
        number.add_argument(
            "-n", "--number", metavar="<number>", type=number_type, default=1,
            help="number of next actions to show (default: %(default)s)")

    def add_filter_arguments(self) -> None:
        """Add the filter arguments to the parser."""
        filters = self.add_argument_group("Limit the tasks from which the next actions are selected")
        # List contexts or projects in the current todo.txt file(s), for tab completion
        filters.add_argument(
            "--list-arguments", help=argparse.SUPPRESS)
        date = filters.add_mutually_exclusive_group()
        date.add_argument(
            "-d", "--due", metavar="<due date>", type=date_type, nargs="?", const=datetime.date.max,
            help="show only next actions with a due date; if a date is given, show only next actions due on or "
                 "before that date")
        date.add_argument("-o", "--overdue", help="show only overdue next actions", action="store_true")
        filters.add_argument(
            "-p", "--priority", metavar="<priority>", choices=string.ascii_uppercase, nargs="?",
            help="minimum priority (A-Z) of next actions to show (default: %(default)s)")
        # Collect all context and project arguments in one list:
        filters.add_argument(
            "filters", metavar="<context|project>", help=argparse.SUPPRESS, nargs="*", type=filter_type)
        filters.add_argument(
            "contexts", metavar="@<context> ...", nargs="*", type=filter_type,
            help="contexts the next action must have")
        filters.add_argument(
            "projects", metavar="+<project> ...", nargs="*", type=filter_type,
            help="projects the next action must be part of; if repeated the next action must be part of at least one "
                 "of the projects")
        filters.add_argument(
            "excluded_contexts", metavar="-@<context> ...", nargs="*", type=filter_type,
            help="contexts the next action must not have")
        filters.add_argument(
            "excluded_projects", metavar="-+<project> ...", nargs="*", type=filter_type,
            help="projects the next action must not be part of")

    def parse_args(self, args=None, namespace=None) -> argparse.Namespace:
        """Parse the command-line arguments."""
        namespace, remaining_args = self.parse_known_args(args, namespace)
        self.parse_remaining_args(remaining_args, namespace)
        namespace.contexts = subset(namespace.filters, "@")
        namespace.projects = subset(namespace.filters, "+")
        namespace.excluded_contexts = subset(namespace.filters, "-@")
        namespace.excluded_projects = subset(namespace.filters, "-+")
        self.validate_arguments(namespace)
        if namespace.time_travel and namespace.due:
            # Apply time travel to options that take a date argument (which currently is only --due)
            namespace.due += namespace.time_travel - datetime.date.today()
        if getattr(namespace, "config_file", self.get_default("config_file")) is not None:
            self.process_config_file(namespace)
        self.fix_filenames(namespace)
        if namespace.write_config_file:
            write_config_file(namespace)
            self.exit()
        return namespace

    def parse_remaining_args(self, args, namespace: argparse.Namespace) -> None:
        """Parse the remaining command-line arguments, i.e. the excluded contexts and projects."""
        try:
            namespace.filters.extend([filter_type(arg) for arg in args])
        except argparse.ArgumentTypeError as reason:
            self.error(str(reason))

    def validate_arguments(self, namespace: argparse.Namespace) -> None:
        """Validate arguments."""
        if any(value == "" for value in namespace.contexts | namespace.excluded_contexts):
            self.error("argument <context|project>: context name missing")
        if any(value == "" for value in namespace.projects | namespace.excluded_projects):
            self.error("argument <context|project>: project name missing")
        for value in namespace.contexts:
            if value in namespace.excluded_contexts:
                self.error(f"@{value} is both included and excluded")
        for value in namespace.projects:
            if value in namespace.excluded_projects:
                self.error(f"+{value} is both included and excluded")

    def process_config_file(self, namespace: argparse.Namespace) -> None:
        """Process the configuration file."""
        config_filename = namespace.config_file
        config = read_config_file(config_filename, self.get_default("config_file"), self.error)
        if not config:
            return
        validate_config_file(config, config_filename, self.error)
        self.insert_config(config, namespace)

    def insert_config(self, config, namespace: argparse.Namespace) -> None:
        """Insert the configured parameters in the namespace, if no command line arguments are present."""
        if self.arguments_not_specified("-f", "--file"):
            filenames = config.get("file", [])
            if isinstance(filenames, str):
                filenames = [filenames]
            getattr(namespace, "file").extend(filenames)
        if self.arguments_not_specified("-n", "--number", "-a", "--all"):
            number = sys.maxsize if config.get("all", False) else config.get("number", self.get_default("number"))
            setattr(namespace, "number", number)
        if self.arguments_not_specified("-r", "--reference"):
            reference = config.get("reference", self.get_default("reference"))
            setattr(namespace, "reference", reference)
        if self.arguments_not_specified("-s", "--style"):
            style = config.get("style", self.get_default("style"))
            setattr(namespace, "style", style)
        if self.arguments_not_specified("-p", "--priority"):
            priority = config.get("priority", self.get_default("priority"))
            setattr(namespace, "priority", priority)
        if self.arguments_not_specified("-b", "--blocked"):
            blocked = config.get("blocked", self.get_default("blocked"))
            setattr(namespace, "blocked", blocked)
        self.insert_configured_filters(config, namespace)

    @staticmethod
    def insert_configured_filters(config, namespace: argparse.Namespace) -> None:
        """Insert the configured filters in the namespace, if no matching command line filters are present."""
        filters = config.get("filters", [])
        if isinstance(filters, str):
            import re
            filters = re.split(r"\s", filters)
        for configured_filter in filters:
            if configured_filter.startswith("@") and configured_filter[len("@"):] not in namespace.excluded_contexts:
                namespace.contexts.add(configured_filter[len("@"):])
            if configured_filter.startswith("+") and configured_filter[len("+"):] not in namespace.excluded_projects:
                namespace.projects.add(configured_filter[len("+"):])
            if configured_filter.startswith("-@") and configured_filter[len("-@"):] not in namespace.contexts:
                namespace.excluded_contexts.add(configured_filter[len("-@"):])
            if configured_filter.startswith("-+") and configured_filter[len("-+"):] not in namespace.projects:
                namespace.excluded_projects.add(configured_filter[len("-+"):])

    @staticmethod
    def arguments_not_specified(*arguments: str) -> bool:
        """Return whether any of the arguments was specified on the command line."""
        return not any([command_line_arg.startswith(argument) for argument in arguments
                        for command_line_arg in sys.argv])

    def fix_filenames(self, namespace: argparse.Namespace) -> None:
        """Fix the filenames."""
        # Work around the issue that the "append" action doesn't overwrite defaults.
        # See https://bugs.python.org/issue16399.
        filenames = namespace.file[:]
        default_filenames = self.__default_filenames
        if default_filenames != filenames:
            for default_filename in default_filenames:
                filenames.remove(default_filename)
        # Remove duplicate filenames while maintaining order.
        namespace.file = list(dict.fromkeys(filenames))


class CapitalisedHelpFormatter(argparse.HelpFormatter):
    """Capitalise the usage string."""

    def add_usage(self, usage, actions, groups, prefix=None):
        """Insert a capitalised usage string."""
        return super().add_usage(usage, actions, groups, prefix or "Usage: ")


def filter_type(value: str) -> str:
    """Return the filter if it's valid, else raise an error."""
    if value.startswith("@") or value.startswith("+") or value.startswith("-@") or value.startswith("-+"):
        return value
    raise argparse.ArgumentTypeError(f"unrecognized argument: {value}")


def date_type(value: str) -> datetime.date:
    """Return the date if it's valid, else raise an error."""
    relative_days = dict(yesterday=-1, today=0, tomorrow=1)
    if value.lower() in relative_days:
        return datetime.date.today() + datetime.timedelta(days=relative_days[value.lower()])
    from dateutil.parser import parse
    try:
        date_time, remaining_tokens = cast(Tuple, parse(value, fuzzy_with_tokens=True, ignoretz=True))
        if not remaining_tokens:
            return date_time.date()
    except ValueError:
        pass
    raise argparse.ArgumentTypeError(f"invalid date: {value}")


def number_type(value: str) -> int:
    """Return the value if it's positive, else raise an error."""
    try:
        number = int(value)
        if number > 0:
            return number
    except ValueError:
        pass
    raise argparse.ArgumentTypeError(f"invalid number: {value}")


def subset(filters: List[str], prefix: str) -> Set[str]:
    """Return a subset of the filters based on prefix."""
    return set(f[len(prefix):] for f in filters if f.startswith(prefix))



"""Unit tests for the Tasks class."""

import unittest

from next_action.todotxt import Tasks, Task


class ContextsTest(unittest.TestCase):
    """Unit tests for the contexts methods."""

    def test_no_tasks(self):
        """Test that an empty tasks collection has no contexts."""
        self.assertEqual(set(), Tasks().contexts())

    def test_task_without_context(self):
        """Test that a task collection with tasks without contexts has not contexts."""
        self.assertEqual(set(), Tasks([Task("Todo")]).contexts())

    def test_task_with_context(self):
        """Test that a task collection with a task with a context has that context."""
        self.assertEqual(set(["work"]), Tasks([Task("Todo @work")]).contexts())

    def test_task_with_contexts(self):
        """Test that a task collection with a task with contexts has those contexts."""
        self.assertEqual(set(["office", "work"]), Tasks([Task("Todo @work @office")]).contexts())

    def test_tasks_with_contexts(self):
        """Test that a task collection with tasks with contexts has those contexts."""
        self.assertEqual(set(["office", "work", "home"]),
                         Tasks([Task("Todo @work @office"), Task("Todo @home @office")]).contexts())


class ProjectsTest(unittest.TestCase):
    """Unit tests for the projects method."""

    def test_no_tasks(self):
        """Test that an empty tasks collection has no projects."""
        self.assertEqual(set(), Tasks().projects())

    def test_task_without_project(self):
        """Test that a task collection with tasks without projects has not projects."""
        self.assertEqual(set(), Tasks([Task("Todo")]).projects())

    def test_task_with_project(self):
        """Test that a task collection with a task with a project has that project."""
        self.assertEqual({"BigProject"}, Tasks([Task("Todo +BigProject")]).projects())

    def test_task_with_projects(self):
        """Test that a task collection with a task with projects has those projects."""
        self.assertEqual({"This", "That"}, Tasks([Task("Todo +This +That")]).projects())

    def test_tasks_with_projects(self):
        """Test that a task collection with tasks with projects has those projects."""
        self.assertEqual({"Foo", "Bar", "Baz"}, Tasks([Task("Todo +Foo +Bar"), Task("Todo +Foo +Baz")]).projects())


class PrioritiesTest(unittest.TestCase):
    """Unit tests for the priorities method."""

    def test_no_tasks(self):
        """Test that an empty tasks collection has no priorities."""
        self.assertEqual(set(), Tasks().priorities())

    def test_task_without_priority(self):
        """Test that a task collection with tasks without priorities has not priorities."""
        self.assertEqual(set(), Tasks([Task("Todo")]).priorities())

    def test_task_with_priority(self):
        """Test that a task collection with a task with a priority has that priority."""
        self.assertEqual(set(["C"]), Tasks([Task("(C) Todo")]).priorities())

    def test_tasks_with_priorities(self):
        """Test that a task collection with tasks with priorities has those priorities."""
        self.assertEqual(set(["X", "Y"]), Tasks([Task("(X) Todo"), Task("(Y) Todo")]).priorities())



"""Create a random todo.txt file for testing purposes."""

import random
import string


def random_bool():
    """Random True/False choice."""
    return random.choice([True, False])  # nosec


def random_string():
    """Return a random string."""
    return random.choice(string.ascii_lowercase)  # nosec


def context():
    """Return a random context."""
    return " @" + random_string() if random_bool() else ""


def project():
    """Return a random project."""
    return " +" + random_string() if random_bool() else ""


def prio():
    """Return a random priority."""
    return f"({random.choice(string.ascii_uppercase)}) " if random_bool() else ""  # nosec


def due_date():
    """Return a random due date."""
    return f" due:2018-{random.randint(1, 12)}-1" if random_bool() else ""  # nosec


def task_id(index):
    """Return a task identifier for 50% of the tasks."""
    return f" id:task{index}" if random_bool() else ""


def create_random_todo_txt():
    """Create a Todo.txt file with random tasks."""
    ids = []
    with open("todo.txt", "w") as todo_txt:
        for i in range(10000):
            before = after = ""
            if len(ids) > 3:
                sample = random.sample(ids, 3)
                while sample and random_bool() and random_bool():
                    before += f" before:task{sample.pop()}" if random_bool() else ""
            id_ = task_id(i)
            todo_txt.write(f"{prio()}Task {i}{id_}{due_date()}{before}{after}{context()}{project()}\n")
            if id_:
                ids.append(i)


if __name__ == "__main__":
    create_random_todo_txt()



"""Unit tests for the warning module in the output package."""

from next_action import todotxt
from next_action.output.warning import invalid_arguments

from .. import fixtures


class InvalidArgumentsTest(fixtures.TestCaseWithNamespace):
    """Unit tests for the invalid argument warning method."""

    def setUp(self):
        """Set up the namespace with default arguments for all unit tests."""
        super().setUp()
        self.tasks = todotxt.Tasks()

    def test_no_args_and_no_tasks(self):
        """Test that the warning is empty if there are no tasks and no arguments."""
        self.assertEqual("", invalid_arguments(self.namespace, self.tasks))

    def test_missing_context(self):
        """Test the message with one missing context."""
        self.namespace.contexts = set(["home"])
        self.assertEqual(" (warning: unknown context: home)", invalid_arguments(self.namespace, self.tasks))

    def test_missing_contexts(self):
        """Test the message with two missing contexts."""
        self.namespace.contexts = set(["home", "work"])
        self.assertEqual(" (warning: unknown contexts: home, work)", invalid_arguments(self.namespace, self.tasks))

    def test_missing_project(self):
        """Test the message with one missing project."""
        self.namespace.projects = set(["DogHouse"])
        self.assertEqual(" (warning: unknown project: DogHouse)", invalid_arguments(self.namespace, self.tasks))

    def test_missing_projects(self):
        """Test the message with two missing projects."""
        self.namespace.projects = set(["DogHouse", "BigTrip"])
        self.assertEqual(" (warning: unknown projects: BigTrip, DogHouse)",
                         invalid_arguments(self.namespace, self.tasks))

    def test_missing_excluded_context(self):
        """Test the message with one missing excluded context."""
        self.namespace.excluded_contexts = set(["home"])
        self.assertEqual(" (warning: unknown context: home)", invalid_arguments(self.namespace, self.tasks))

    def test_missing_excluded_project(self):
        """Test the message with one missing excluded project."""
        self.namespace.excluded_projects = set(["DogHouse"])
        self.assertEqual(" (warning: unknown project: DogHouse)", invalid_arguments(self.namespace, self.tasks))

    def test_missing_context_project(self):
        """Test the message with one missing context and one missing project."""
        self.namespace.contexts = set(["home"])
        self.namespace.projects = set(["DogHouse"])
        self.assertEqual(" (warning: unknown context: home; unknown project: DogHouse)",
                         invalid_arguments(self.namespace, self.tasks))

    def test_present_context(self):
        """Test the message with one present context."""
        self.namespace.contexts = set(["home"])
        self.tasks.append(todotxt.Task("A task @home"))  # pylint: disable=no-member
        self.assertEqual("", invalid_arguments(self.namespace, self.tasks))



"""Unit test for the next action algorithm."""

import datetime

from next_action import todotxt, pick_action

from . import fixtures


class PickActionTest(fixtures.TestCaseWithNamespace):
    """Unit tests for the pick action algorithm."""

    def test_no_tasks(self):
        """Test that no tasks means no next action."""
        self.assertEqual([], pick_action.next_actions([], self.namespace))

    def test_one_task(self):
        """If there is one task, that one is the next action."""
        task = todotxt.Task("Todo")
        self.assertEqual([task], pick_action.next_actions([task], self.namespace))

    def test_multiple_tasks(self):
        """If there are multiple tasks, the first is the next action."""
        task1 = todotxt.Task("Todo 1")
        task2 = todotxt.Task("Todo 2")
        self.assertEqual([task1, task2], pick_action.next_actions([task1, task2], self.namespace))

    def test_higher_prio_goes_first(self):
        """Test that a task with higher priority takes precedence."""
        task1 = todotxt.Task("Todo 1")
        task2 = todotxt.Task("(B) Todo 2")
        task3 = todotxt.Task("(A) Todo 3")
        self.assertEqual([task3, task2, task1], pick_action.next_actions([task1, task2, task3], self.namespace))

    def test_creation_dates(self):
        """Test that a task with an older creation date takes precedence."""
        no_creation_date = todotxt.Task("Task 1")
        newer_task = todotxt.Task("2018-02-02 Task 2")
        older_task = todotxt.Task("2017-01-01 Task 3")
        self.assertEqual([older_task, newer_task, no_creation_date],
                         pick_action.next_actions([no_creation_date, newer_task, older_task], self.namespace))

    def test_priority_and_creation_date(self):
        """Test that priority takes precedence over creation date."""
        priority = todotxt.Task("(C) Task 1")
        newer_task = todotxt.Task("2018-02-02 Task 2")
        older_task = todotxt.Task("2017-01-01 Task 3")
        self.assertEqual([priority, older_task, newer_task],
                         pick_action.next_actions([priority, newer_task, older_task], self.namespace))

    def test_due_dates(self):
        """Test that a task with an earlier due date takes precedence."""
        no_due_date = todotxt.Task("Task 1")
        earlier_task = todotxt.Task("Task 2 due:2018-02-02")
        later_task = todotxt.Task("Task 3 due:2019-01-01")
        self.assertEqual([earlier_task, later_task, no_due_date],
                         pick_action.next_actions([no_due_date, later_task, earlier_task], self.namespace))

    def test_due_and_creation_dates(self):
        """Test that a task with a due date takes precedence over creation date."""
        task1 = todotxt.Task("2018-1-1 Task 1")
        task2 = todotxt.Task("Task 2 due:2018-1-1")
        self.assertEqual([task2, task1], pick_action.next_actions([task1, task2], self.namespace))

    def test_project(self):
        """Test that a task with a project takes precedence over a task without a project."""
        task1 = todotxt.Task("Task 1")
        task2 = todotxt.Task("Task 2 +Project")
        self.assertEqual([task2, task1], pick_action.next_actions([task1, task2], self.namespace))

    def test_projects(self):
        """Test that a task with more projects takes precedence over a task with less projects."""
        task1 = todotxt.Task("Task 1 +Project")
        task2 = todotxt.Task("Task 2 +Project1 +Project2")
        self.assertEqual([task2, task1], pick_action.next_actions([task1, task2], self.namespace))


class FilterTasksTest(fixtures.TestCaseWithNamespace):
    """Test that the tasks from which the next action is picked, can be filtered."""

    def test_context(self):
        """Test that the next action can be limited to a specific context."""
        task1 = todotxt.Task("Todo 1 @work @computer")
        task2 = todotxt.Task("(B) Todo 2 @work")
        task3 = todotxt.Task("(A) Todo 3 @home")
        self.namespace.contexts = {"work"}
        self.assertEqual([task2, task1], pick_action.next_actions([task1, task2, task3], self.namespace))

    def test_contexts(self):
        """Test that the next action can be limited to a set of contexts."""
        task1 = todotxt.Task("Todo 1 @work @computer")
        task2 = todotxt.Task("(B) Todo 2 @work @computer")
        task3 = todotxt.Task("(A) Todo 3 @home @computer")
        self.namespace.contexts = {"work", "computer"}
        self.assertEqual([task2, task1], pick_action.next_actions([task1, task2, task3], self.namespace))

    def test_excluded_context(self):
        """Test that contexts can be excluded."""
        task = todotxt.Task("(A) Todo @computer")
        self.namespace.excluded_contexts = {"computer"}
        self.assertEqual([], pick_action.next_actions([task], self.namespace))

    def test_excluded_contexts(self):
        """Test that contexts can be excluded."""
        task = todotxt.Task("(A) Todo @computer @work")
        task = todotxt.Task("(A) Todo @computer")
        self.namespace.excluded_contexts = {"computer"}
        self.assertEqual([], pick_action.next_actions([task], self.namespace))

    def test_not_excluded_context(self):
        """Test that a task is not excluded if it doesn't belong to the excluded category."""
        task = todotxt.Task("(A) Todo @computer")
        self.namespace.excluded_contexts = {"phone"}
        self.assertEqual([task], pick_action.next_actions([task], self.namespace))

    def test_project(self):
        """Test that the next action can be limited to a specific project."""
        task1 = todotxt.Task("Todo 1 +ProjectX")
        task2 = todotxt.Task("(B) Todo 2 +ProjectX")
        task3 = todotxt.Task("(A) Todo 3 +ProjectY")
        self.namespace.projects = {"ProjectX"}
        self.assertEqual([task2, task1], pick_action.next_actions([task1, task2, task3], self.namespace))

    def test_excluded_project(self):
        """Test that projects can be excluded."""
        task = todotxt.Task("(A) Todo +DogHouse")
        self.namespace.excluded_projects = {"DogHouse"}
        self.assertEqual([], pick_action.next_actions([task], self.namespace))

    def test_excluded_projects(self):
        """Test that projects can be excluded."""
        task = todotxt.Task("(A) Todo +DogHouse +PaintHouse")
        self.namespace.excluded_projects = {"DogHouse"}
        self.assertEqual([], pick_action.next_actions([task], self.namespace))

    def test_not_excluded_project(self):
        """Test that a task is not excluded if it doesn't belong to the excluded project."""
        task = todotxt.Task("(A) Todo +DogHouse")
        self.namespace.excluded_projects = {"PaintHouse"}
        self.assertEqual([task], pick_action.next_actions([task], self.namespace))

    def test_project_and_context(self):
        """Test that the next action can be limited to a specific project and context."""
        task1 = todotxt.Task("Todo 1 +ProjectX @office")
        task2 = todotxt.Task("(B) Todo 2 +ProjectX")
        task3 = todotxt.Task("(A) Todo 3 +ProjectY")
        self.namespace.contexts = {"office"}
        self.namespace.projects = {"ProjectX"}
        self.assertEqual([task1], pick_action.next_actions([task1, task2, task3], self.namespace))


class IgnoredTasksTest(fixtures.TestCaseWithNamespace):
    """Test that certain tasks are ignored when picking the next action."""

    def test_ignore_future_task(self):
        """Ignore tasks with a start date in the future."""
        future_task = todotxt.Task("(A) 9999-01-01 Start preparing for five-digit years")
        regular_task = todotxt.Task("(B) Look busy")
        self.assertEqual([regular_task], pick_action.next_actions([future_task, regular_task], self.namespace))

    def test_only_future_tasks(self):
        """If all tasks are future tasks, there's no next action."""
        future_task1 = todotxt.Task("(A) 9999-01-01 Start preparing for five-digit years")
        future_task2 = todotxt.Task("(A) Start preparing for five-digit years t:9999-01-01")
        self.assertEqual([], pick_action.next_actions([future_task1, future_task2], self.namespace))


class OverdueTasks(fixtures.TestCaseWithNamespace):
    """Unit tests for the overdue filter."""

    def test_overdue_tasks(self):
        """Test that tasks that not overdue are filtered."""
        no_duedate = todotxt.Task("Task")
        future_duedate = todotxt.Task("Task due:9999-01-01")
        overdue = todotxt.Task("Task due:2000-01-01")
        self.namespace.overdue = True
        self.namespace.time_travel = None
        self.assertEqual([overdue], pick_action.next_actions([no_duedate, future_duedate, overdue], self.namespace))


class DueTasks(fixtures.TestCaseWithNamespace):
    """Unit tests for the is_due filter."""

    def test_due_tasks(self):
        """Test that tasks that are not due are filtered."""
        no_duedate = todotxt.Task("Task")
        future_duedate = todotxt.Task("Task due:9999-01-01")
        overdue = todotxt.Task("Task due:2000-01-01")
        self.namespace.due = datetime.date(2000, 1, 1)
        self.assertEqual([overdue], pick_action.next_actions([no_duedate, future_duedate, overdue], self.namespace))

    def test_any_due_tasks(self):
        """Test that tasks that are not due are filtered."""
        no_duedate = todotxt.Task("Task")
        future_duedate = todotxt.Task("Task due:9999-01-01")
        overdue = todotxt.Task("Task due:2000-01-01")
        self.namespace.due = datetime.date.max
        self.assertEqual([overdue, future_duedate],
                         pick_action.next_actions([no_duedate, future_duedate, overdue], self.namespace))


class MinimimPriorityTest(fixtures.TestCaseWithNamespace):
    """Unit test for the mininum priority filter."""

    def test_priority(self):
        """Test that tasks without priority are filtered."""
        no_priority = todotxt.Task("Task")
        high_priority = todotxt.Task("(A) Task")
        low_priority = todotxt.Task("(Z) Task")
        self.namespace.priority = "K"
        self.assertEqual(
            [high_priority], pick_action.next_actions([no_priority, low_priority, high_priority], self.namespace))


class BlockedTasksTest(fixtures.TestCaseWithNamespace):
    """Unit tests for blocked tasks."""

    # pylint: disable=no-member

    def test_due_date_of_blocked_task(self):
        """Test that a task that blocks a task with an earlier due date takes precendence."""
        tasks = todotxt.Tasks()
        due_first_but_blocked = todotxt.Task("Task id:1 due:2018-01-01")
        blocking_task = todotxt.Task("Blocking before:1")
        due_second = todotxt.Task("Task due:2018-02-01")
        due_first_but_blocked.set_is_blocked()
        blocking_task.add_blocked_task(due_first_but_blocked)
        tasks.extend([blocking_task, due_second])
        self.assertEqual([blocking_task, due_second], pick_action.next_actions(tasks, self.namespace))

    def test_priority_of_blocked_task(self):
        """Test that a task that blocks a task with a higher priority takes precendence."""
        high_prio_but_blocked = todotxt.Task("(A) Task id:1")
        blocking_task = todotxt.Task("Blocking before:1")
        second_prio = todotxt.Task("(B) Task")
        high_prio_but_blocked.set_is_blocked()
        blocking_task.add_blocked_task(high_prio_but_blocked)
        tasks = todotxt.Tasks([second_prio, blocking_task])
        self.assertEqual([blocking_task, second_prio], pick_action.next_actions(tasks, self.namespace))



"""Module for warning messages."""

import argparse

from . import todotxt


def invalid_arguments(namespace: argparse.Namespace, tasks: todotxt.Tasks) -> str:
    """Check whether the context and projects given on the command line actually exist in the task file."""
    unknown_contexts = (namespace.contexts | namespace.excluded_contexts) - tasks.contexts()
    unknown_projects = (namespace.projects | namespace.excluded_projects) - tasks.projects()
    error_messages = []
    if unknown_contexts:
        plural = "s" if len(unknown_contexts) > 1 else ""
        error_messages.append(f"unknown context{plural}: {', '.join(sorted(unknown_contexts))}")
    if unknown_projects:
        plural = "s" if len(unknown_projects) > 1 else ""
        error_messages.append(f"unknown project{plural}: {', '.join(sorted(unknown_projects))}")
    return f" (warning: {'; '.join(error_messages)})" if error_messages else ""



"""Algorithm for deciding the next action(s)."""

import argparse
import datetime

from typing import Tuple

from . import todotxt


def sort_key(task: todotxt.Task) -> Tuple[str, datetime.date, datetime.date, int]:
    """Return the sort key for a task."""
    return (task.priority() or "ZZZ", task.due_date() or datetime.date.max, task.creation_date() or datetime.date.max,
            -len(task.projects()))


def next_actions(tasks: todotxt.Tasks, arguments: argparse.Namespace) -> todotxt.Tasks:
    """Return the next action(s) from the collection of tasks."""
    contexts = arguments.contexts
    projects = arguments.projects
    excluded_contexts = arguments.excluded_contexts
    excluded_projects = arguments.excluded_projects

    # First, get the potential next actions by filtering out tasks with a future creation date or threshold date
    time_travel_date = arguments.time_travel or datetime.date.today()
    eligible_tasks = filter(lambda task: not task.is_future(time_travel_date), tasks)
    # Then, exclude tasks that have an excluded context
    if excluded_contexts:
        eligible_tasks = filter(lambda task: not excluded_contexts & task.contexts(), eligible_tasks)
    # And, tasks that have an excluded project
    if excluded_projects:
        eligible_tasks = filter(lambda task: not excluded_projects & task.projects(), eligible_tasks)
    # Then, select the tasks that belong to all given contexts, if any
    if contexts:
        eligible_tasks = filter(lambda task: contexts <= task.contexts(), eligible_tasks)
    # Next, select the tasks that belong to at least one of the given projects, if any
    if projects:
        eligible_tasks = filter(lambda task: projects & task.projects(), eligible_tasks)
    # If the user only wants to see overdue tasks, filter out non-overdue tasks
    if arguments.overdue:
        eligible_tasks = filter(lambda task: task.is_overdue(time_travel_date), eligible_tasks)
    # If the user only wants to see tasks due before a due date, filter out non-due tasks
    if arguments.due:
        eligible_tasks = filter(lambda task: task.is_due(arguments.due), eligible_tasks)
    # If the user specified a minimum priority, filter out tasks with a lower priority or no priority
    if arguments.priority:
        eligible_tasks = filter(lambda task: task.priority_at_least(arguments.priority), eligible_tasks)
    # Finally, sort by priority, due date and creation date
    return todotxt.Tasks(sorted(eligible_tasks, key=sort_key)[:arguments.number])
